import collections
import logging
import re
import sys
import textwrap

from . import options

Token = collections.namedtuple('Token', ['kind', 'value'])
sys.setrecursionlimit(100000)

def lexer(text):
    __tokens = [
        ('COMMENT', ';[^\n]*'),
        ('LPAREN', '\\('),
        ('RPAREN', '\\)'),
        ('STRINGLIT', '"[^"]*"'),
        ('QUOTEDSYM', '\\|[^\\|]*\\|'),
        ('SYMBOL', '[:a-zA-Z0-9~!@$#%\\^&*_+=<>.?/-]+'),
        ('SPACE', '\\s+'),
        ('MISMATCH', '.'),
    ]
    __token_re = re.compile('|'.join(['(?P<{}>{})'.format(tok[0], tok[1]) for tok in __tokens]))

    for m in __token_re.finditer(text):
        kind = m.lastgroup
        value = m.group()
        if kind in ['COMMENT', 'SPACE']:
            continue
        if kind == 'MISMATCH':
            logging.warning('Unexpected  {}'.format(value))
        yield Token(kind, value)

class Peekable:
    def __init__(self, generator):
        self.__gen = generator
        self.__peek = None
        self.__empty = False
        try:
            self.__peek = next(self.__gen)
        except StopIteration:
            self.__empty = True

    def __iter__(self):
        return self
    def __next__(self):
        if self.__empty:
            raise StopIteration
        res = self.__peek
        try:
            self.__peek = next(self.__gen)
        except StopIteration:
            self.__peek = None
            self.__empty = True
        return res
    def empty(self):
        return self.__empty
    def peek(self):
        return self.__peek


def parse_expression(tokens):
    tok = next(tokens)
    if tok.kind == 'LPAREN':
        args = []
        while tokens.peek().kind != 'RPAREN':
            args.append(parse_expression(tokens))
        assert tokens.peek().kind == 'RPAREN'
        next(tokens)
        return args
    return tok.value

def parse_smtlib(text):
    """Parses an SMT-LIB input to a sequence of nodes."""
    token_stream = Peekable(lexer(text))
    exprs = []
    while not token_stream.empty():
        exprs.append(parse_expression(token_stream))
    return exprs

def render_expression(expr):
    """Renders a node to a string."""
    if isinstance(expr, list):
        return '(' + ' '.join(map(render_expression, expr)) + ')'
    return expr

def render_pretty_expression(expr, indent = ''):
    """Renders a node to a string in a pretty way."""
    if isinstance(expr, list):
        if expr != [] and expr[0] in ['declare-const', 'declare-fun']:
            return render_expression(expr)
        if all(map(lambda e: not isinstance(e, list), expr)):
            return '(' + ' '.join(expr) + ')'
        res = '(' + render_pretty_expression(expr[0]) + '\n'
        for e in expr[1:]:
            res += indent + '\t' + render_pretty_expression(e, indent + '\t') + '\n'
        res += indent + ')'
        return res
    return expr

def render_smtlib(exprs):
    """Renders a sequence of nodes to a string."""
    if options.args().pretty_print:
        return '\n'.join(map(render_pretty_expression, exprs))
    res = map(render_expression, exprs)
    if options.args().wrap_lines:
        res = [line for r in res for line in textwrap.wrap(r, width = 78, subsequent_indent = '  ')]
    return '\n'.join(res)

def write_smtlib_to_file(exprs, filename):
    """Writes a sequence of nodes to a file."""
    open(filename, 'w').write(render_smtlib(exprs))



import logging
import os
import pprint
import sys
import time

from pydelta import checker
from pydelta import options
from pydelta import manager
from pydelta import mutator
from pydelta import parser

def setup_logging():
    logging.basicConfig(format = '[pydelta %(levelname)s] %(message)s')
    if options.args().verbose:
        logging.getLogger().setLevel(level=logging.DEBUG)
    else:
        logging.getLogger().setLevel(level=logging.INFO)

def check_options():
    if options.args().max_threads != 1:
        # configure number of threads
        if options.args().max_threads <= 0:
            options.args().max_threads = os.cpu_count() + options.args().max_threads
        logging.info('Using up to %d threads.', options.args().max_threads)

    if options.args().dump_config:
        pprint.pprint(vars(options.args()))

    # check input file
    if not os.path.isfile(options.args().inputfile):
        raise Exception('Input file is not a regular file')

    if options.args().parse_only:
        # only parse and print
        exprs = parser.parse_smtlib(open(options.args().inputfile).read())
        pprint.pprint(exprs)
        sys.exit(0)

    # check executable
    if len(options.args().cmd) == 0:
        raise Exception('No executable was specified as command')
    if not os.path.isfile(options.args().cmd[0]):
        raise Exception('Command "{}" is not a regular file'.format(options.args().cmd[0]))
    if not os.access(options.args().cmd[0], os.X_OK):
        raise Exception('Command "{}" is not executable'.format(options.args().cmd[0]))

def do_reference_run():
    if not checker.compute_reference(options.args().cmd, options.args().inputfile):
        logging.error('Computing the reference output failed.')
        sys.exit(1)

def parse_input():
    return parser.parse_smtlib(open(options.args().inputfile).read())

def run_pydelta(original):
    mutator.collect_mutators(options.args())
    skip = options.args().skip
    simplifications = 0
    m = manager.Manager()
    while True:
        try:
            # do one simplification step
            start = time.time()
            simp = m.simplify(original, skip)
            duration = time.time() - start
        except KeyboardInterrupt:
            logging.warning('Aborting. See %s for results.', options.args().outputfile)
            break
        if simp is None:
            logging.info('No further simplification found')
            if skip > 0:
                skip = 0
                logging.info('Starting over')
            else:
                # terminate
                parser.write_smtlib_to_file(original, options.args().outputfile)
                logging.info('Final input (written to %s):\n%s', options.args().outputfile, parser.render_smtlib(original))
                break
        else:
            simplifications += 1
            # write current status to file and continue
            logging.info('#%d: %s (%.2fs)', simplifications, simp.simplification, duration)
            skip = simp.counter
            original = simp.exprs
            parser.write_smtlib_to_file(original, options.args().outputfile)

    logging.info('Performed %d checks and %d simplifications', checker.CHECKS, simplifications)
    if checker.TIMEOUTS > 0:
        logging.info('Overall, checking timed out %d times', checker.TIMEOUTS)



from . import options
from .semantics import *

NAME = 'core'
MUTATORS = [
    'constants', 'erase-children', 'inline-functions', 'let-elimination', 'let-substitution',
    'merge-children', 'replace-by-variable', 'sort-children', 'substitute-children'
]

class Constants:
    """Replaces any node by a constant."""
    def mutations(self, node):
        """Return :code:`get_constants(get_return_type(node))`."""
        res = get_constants(get_return_type(node))
        if node in res:
            return []
        return res
    def __str__(self):
        return 'substitute by a constant'

class EraseChildren:
    """Erases a single child of the given node."""
    def filter(self, node):
        return not is_leaf(node)
    def mutations(self, node):
        res = []
        for i in range(len(node)):
            tmp = node.copy()
            del tmp[i]
            res.append(tmp)
        return res
    def __str__(self):
        return 'erase child'

class TopLevelBinaryReduction:
    """Performs binary reduction on the top level node. Essentially mimics line based reduction."""
    def global_mutations(self, linput, ginput):
        if linput != ginput:
            return []
        # generate all sublists as generated by binary-search in bfs order
        # let den be the denominator of the list length (the tree level)
        # let num be the numerator within the current tree level
        res = []
        den = 2
        while den < len(ginput):
            for num in range(0, den):
                start = int(num/den * len(ginput))
                end = int((num+1)/den * len(ginput))
                res.append(ginput[:start] + ginput[end:])
            den *= 2
        return res
    def __str__(self):
        return 'binary reduction'

class MergeWithChildren:
    """Merges a node with one of its children. This is possible for n-ary operators like :code:`and` or :code:`+`."""
    def filter(self, node):
        return is_nary(node)
    def mutations(self, node):
        res = []
        for cid, child in enumerate(node):
            if has_name(child) and get_name(node) == get_name(child):
                res.append(node[:cid] + node[cid][1:] + node[cid + 1:])
        return res
    def __str__(self):
        return 'merge with child'

class ReplaceByVariable:
    """Replaces a node by a variable."""
    def filter(self, node):
        return not is_constant(node) and has_type(node)
    def mutations(self, node):
        if is_leaf(node):
            variables = get_variables_with_type(get_type(node))
            variables = get_variable_info().keys()
            if options.args().replace_by_variable_mode == 'inc':
                return [v for v in variables if v > node]
            return [v for v in variables if v < node]
        return [v for v in get_variable_info().keys() if node_count(v) < node_count(node)]
    def __str__(self):
        return 'substitute by existing variable'

class SortChildren:
    """Sorts the children of a node."""
    def filter(self, node):
        return not is_leaf(node)
    def mutations(self, node):
        s = sorted(node, key = node_count)
        if s != node:
            return [s]
        return []
    def __str__(self):
        return 'sort children'

class SubstituteChildren:
    """Substitutes a node with one of its children."""
    def filter(self, node):
        return not is_leaf(node) and not is_let(node)
    def mutations(self, node):
        return node[1:]
    def __str__(self):
        return 'substitute with child'

def collect_mutator_options(argparser):
    options.add_mutator_argument(argparser, NAME, True, 'core mutators')
    options.add_mutator_argument(argparser, 'constants', True, 'replace by theory constants')
    options.add_mutator_argument(argparser, 'erase-children', True, 'erase individual children of nodes')
    options.add_mutator_argument(argparser, 'merge-children', True, 'merge children into nodes')
    options.add_mutator_argument(argparser, 'replace-by-variable', True, 'replace with existing variable')
    argparser.add_argument('--replace-by-variable-mode',
                           choices = ['inc', 'dec'], default = 'inc', help = 'replace with existing variables that are larger or smaller')
    options.add_mutator_argument(argparser, 'sort-children', True, 'sort children of nodes')
    options.add_mutator_argument(argparser, 'substitute-children', True, 'substitute nodes with their children')
    options.add_mutator_argument(argparser, 'top-level-binary-reduction', True, 'use top level binary reduction')

def collect_mutators(args):
    res = []
    if args.mutator_core:
        if args.mutator_top_level_binary_reduction:
            res.append(TopLevelBinaryReduction())
        if args.mutator_erase_children:
            res.append(EraseChildren())
        if args.mutator_constants:
            res.append(Constants())
        if args.mutator_merge_children:
            res.append(MergeWithChildren())
        if args.mutator_replace_by_variable:
            res.append(ReplaceByVariable())
        if args.mutator_sort_children:
            res.append(SortChildren())
        if args.mutator_substitute_children:
            res.append(SubstituteChildren())
    return res



import collections
import copy
import logging
import queue
import sys
import tempfile
import threading

from . import checker
from . import mutator
from . import options
from . import parser
from . import semantics

Candidate = collections.namedtuple('Candidate', ['counter', 'simplification', 'exprs'])
"""Represents a simplification candidate.

:code:`counter` contains the number of the simplified node in the pre-order iteration of :meth:`semantics.iterate_nodes`.

:code:`exprs` contains the simplified input.

:code:`simplification` contains the name of the applied mutator.
"""

class Manager:
    """Manages the asynchronous generation and checking of mutated inputs.
    One thread runs the :meth:`producer` method that fills a :class:`queue.Queue`
    while as many threads as given by the :code:`--max-threads` options run and evaluate the candidates from the queue.
    The :meth:`simplify` methods starts all threads and terminates them as soon as one valid simplication has been found.
    """
    def __init__(self):
        self.q = queue.Queue(maxsize = 20)
        self.stop_operation = False
        self.finished_generation = False
        self.result = None
        self.result_lock = threading.Lock()

    def __empty_queue(self):
        """Empty the queue."""
        try:
            while not self.q.empty():
                self.q.get(timeout = 0.1)
                self.q.task_done()
        except queue.Empty:
            pass


    def producer(self, original, skip):
        """Produces new mutated variants of the given input."""
        original_size = semantics.node_count(original)
        gen = mutator.MutationGenerator(skip)
        for candidate in gen.generate_mutations(original):
            if options.args().mode_aggressive:
                if semantics.node_count(candidate[2]) > original_size * (1 - options.args().aggressiveness):
                    continue
            if options.args().mode_reduction_only:
                if semantics.node_count(candidate[2]) >= original_size:
                    continue
            self.q.put(Candidate(candidate[0], candidate[1], copy.deepcopy(candidate[2])))
            if self.stop_operation:
                break
        self.finished_generation = True

    def consumer(self):
        """Takes candidates from the queue and checks whether their output matches the reference result."""
        while not self.stop_operation:
            try:
                candidate = self.q.get(timeout = 0.25)
                self.q.task_done()
                try:
                    with tempfile.NamedTemporaryFile('w', suffix = '.smt2') as tmp:
                        tmp.write(parser.render_smtlib(candidate.exprs))
                        tmp.flush()
                        res = checker.execute(options.args().cmd, tmp.name)
                except FileNotFoundError:
                    logging.info('Removing the temporary file failed.')
                if checker.matches_reference(res):
                    with self.result_lock:
                        if self.result is None:
                            self.stop_operation = True
                            self.result = candidate
            except queue.Empty:
                if self.finished_generation:
                    break
        self.__empty_queue()

    def simplify(self, original, skip = 0):
        """Starts one producer thread and multiple consumer thread and then waits for a valid simplification."""
        assert self.q.empty()
        self.q = queue.Queue(maxsize = 20)
        self.stop_operation = False
        self.finished_generation = False
        self.result = None
        try:
            threads = [
                threading.Thread(target = self.producer, name = 'producer', args = (original, skip))
            ] + [
                threading.Thread(target = self.consumer, name = 'consumer-{}'.format(i + 1))
                for i in range(options.args().max_threads)
            ]
            for t in threads:
                t.start()
            for t in threads:
                t.join()
            self.__empty_queue()
        except KeyboardInterrupt:
            sys.stdout.write('\n')
            logging.warning('Stopping all computations.')
            self.stop_operation = True
            self.__empty_queue()
            raise

        sys.stdout.write('\n')
        return self.result
